#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/delay.h>
#include <linux/pci.h>
#include <linux/interrupt.h>
#include <linux/time.h>
#include <linux/hrtimer.h>
#include <asm/bitops.h>
#include <asm/pci.h>
#include <linux/mm.h>
#include <linux/random.h>
#include <linux/spinlock.h>
#include <linux/pid.h>
#include <linux/sched.h>
#include <linux/sched/signal.h>
#include <linux/uaccess.h>
#include <linux/completion.h>
#include "spc_device.h"

MODULE_AUTHOR("Semen M. Krylov (ksm@ntcms.org)");
MODULE_DESCRIPTION("Xilinx based spc device driver");
MODULE_LICENSE("GPL");

#define UMS7_CLASS "ums7"

struct ums7_data {
	struct spc_device mdev;
	struct pci_dev *pdev;
	void *__iomem io;
	unsigned long phys;
	struct spc_caps caps;
};

static DEFINE_MUTEX(check_reset_mutex);
static unsigned int devices_nr;

static inline struct ums7_data *to_ums7_data(struct spc_device *mdev)
{
	return container_of(mdev, struct ums7_data, mdev);
}

static void ums7_release(struct spc_device *mdev)
{
	(void)mdev;
}

static int ums7_open(struct spc_device *mdev)
{
	if (!try_module_get(THIS_MODULE)) {
		printk(KERN_ERR "%s: open failed\n", UMS7_CLASS);
		return -ENODEV;
	}
	return 0;
}

static void ums7_close(struct spc_device *mdev)
{
	module_put(THIS_MODULE);
}

static void ums7_munmap(struct spc_device *mdev, struct spc_fd *fd)
{
	iounmap(fd->device_mmap_addr);
}

static int ums7_mmap(struct spc_device *mdev, struct vm_area_struct *vma)
{
	unsigned long vsize = vma->vm_end - vma->vm_start;
	struct ums7_data *d = to_ums7_data(mdev);
	int err;

	if (vsize != (128 << 20)) {
		printk(KERN_WARNING "%s: mmap vma size is wrong\n",
				d->caps.name);
		return -EINVAL;
	}

	vma->vm_flags |= VM_IO | VM_DONTEXPAND | VM_DONTDUMP;
	vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);

	err = io_remap_pfn_range(vma, vma->vm_start, vma->vm_pgoff,
			vsize, vma->vm_page_prot);

	if (err)
		printk(KERN_ERR "%s: remap_pfn_range with code %d\n",
				d->caps.name, err);
	return err;
}

static int ums7_check_reset(struct spc_device *mdev,
				unsigned int __user *status);
static struct spc_ops default_ums7_ops = {
	.release = ums7_release,
	.open = ums7_open,
	.close = ums7_close,
	.mmap = ums7_mmap,
	.munmap = ums7_munmap,
	.check_reset = ums7_check_reset,
};

static int ums7_set_device_caps(struct spc_caps *caps,
		const char *name)
{
	snprintf(caps->name, sizeof(caps->name), "%s", name);

	caps->core_nr = 5;
	caps->flags = 0;
	caps->pmr_nr = 1;

	return 0;
}

static int ums7_probe(struct pci_dev *pdev, const struct pci_device_id *ent)
{
	struct ums7_data *dev;
	int ret;

	dev = kzalloc(sizeof(*dev), GFP_DMA);
	if (!dev) {
		printk(KERN_ERR "%s: failed to allocate %lu bytes\n",
				UMS7_CLASS, sizeof(*dev));
		ret = -ENOMEM;
		goto out;
	}
	pci_set_drvdata(pdev, dev);

	if ((ret = pci_enable_device(pdev))) {
		dev_err(&pdev->dev, "%s: pci_enable_device() failed,"
			       " not attaching\n", UMS7_CLASS);
		goto out_reset_drvdata;
	}

	if ((ret = pci_request_regions(pdev, UMS7_CLASS))) {
		dev_err(&pdev->dev, "%s: pci_request_regions() failed,"
				" not attaching\n", UMS7_CLASS);
		goto out_disable_device;
	}

	dev->io = pci_iomap(pdev, 0, 128 << 20);
	if (!dev->io) {
		dev_err(&pdev->dev, "failed to map device io\n");
		ret = -ENXIO;
		goto out_release_regions;
	}

	dev->phys = pci_resource_start(pdev, 0);

	ums7_set_device_caps(&dev->caps, UMS7_CLASS);
	dev->mdev.caps = &dev->caps;
	dev->mdev.caps->flags = SPC_CAP_EXPORT_PMR;
	dev->mdev.caps->pmr_offs[0] = dev->phys;
	printk(KERN_INFO "%s: PCI window at 0x%p\n",
		dev->caps.name, (void *)dev->mdev.caps->pmr_offs[0]);
	dev->mdev.caps->pmr_sizes[0] = 128 << 20;
	dev->mdev.caps->pmr_nr = 1;
	dev->mdev.rev = 5;
	dev->mdev.ops = &default_ums7_ops;

	printk(KERN_INFO "%s: device PCI at 0x%p\n",
			dev->caps.name, (void *)dev->phys);

	snprintf(&dev->mdev.hid.hid[0], sizeof(dev->mdev.hid.hid),
		"SPI:%d:%s:%d", devices_nr++,
		pdev->bus->name, pdev->devfn);

	dev->pdev = pdev;
	dev->mdev.pcid = pdev;

	ret = spc_register(&dev->mdev);
	if (ret) {
		dev_err(&pdev->dev, "failed to register spc device\n");
		goto out_unmap;
	}

	return 0;

out_unmap:
	pci_iounmap(pdev, dev->io);
out_release_regions:
	pci_release_regions(pdev);
out_disable_device:
	pci_disable_device(pdev);
out_reset_drvdata:
	pci_set_drvdata(pdev, 0);
out:
	return ret;
}

static void ums7_remove(struct pci_dev *pdev)
{
	struct ums7_data *dev;

	dev = (struct ums7_data*) pci_get_drvdata(pdev);
	if (!dev) {
		dev_err(&pdev->dev, "%s: device not found in"
				" %s\n", UMS7_CLASS, __FUNCTION__);
		return;
	}

	spc_unregister(&dev->mdev);
	pci_iounmap(pdev, dev->io);
	pci_release_regions(pdev);
	pci_disable_device(pdev);
	pci_set_drvdata(pdev, 0);
}

static int ums7_check_scratch(struct ums7_data *ums_dev)
{
	unsigned int *__iomem io;
	volatile unsigned int in, out;

	io = (unsigned int *__iomem)ums_dev->io;
	get_random_bytes((void*)&in, sizeof(int));
	io[4 >> 2] = in;
	out = io[8 >> 2];
	/* Trick to deceive GCC. */
	get_random_bytes((void *)&out, sizeof(int));
	out = io[4 >> 2];

	if (in == out)
		return 0;

	printk(KERN_ERR "SCRATCH CHECK FAILED! "
			"in = %x, out = %x\n", in, out);

	return 1;
}

static void ums7_reset(struct ums7_data *ums_dev)
{
	unsigned int *__iomem io;

	io = (unsigned int *__iomem)ums_dev->io;

	if (io[0]) {
		io[0x904 >> 2] = 1;
		usleep_range(15,25);
	}

	io[0x100] = 1;
	usleep_range(15,25);
}

static int ums7_check_reset(struct spc_device *mdev,
				unsigned int __user *status)
{
	struct ums7_data *ums_dev = to_ums7_data(mdev);
	int res;

	res = mutex_lock_interruptible(&check_reset_mutex);
	if (res)
		return res;

	res = 0;
	if (ums7_check_scratch(ums_dev)) {
		ums7_reset(ums_dev);
		if (ums7_check_scratch(ums_dev)) {
			res = -EFAULT;
			printk(KERN_ERR "FAILED TO RESTORE PCI DEVICE!!!\n");
		} else {
			printk(KERN_DEBUG "PCI DEVICE RESTORED!!!\n");
			res = 1;
		}
	}

	mutex_unlock(&check_reset_mutex);

	if (copy_to_user(status, &res, sizeof(*status)))
		printk(KERN_DEBUG "FAILED TO SEND INFO TO USER\n");

	return res;
}

#if !defined(DEFINE_PCI_DEVICE_TABLE)
#define DEFINE_PCI_DEVICE_TABLE(name)\
	static struct pci_device_id name[]
#endif

#if !defined(PCI_DEVICE_SUB)
#define PCI_DEVICE_SUB(vend, dev, svend, sdev) \
	.vendor = (vend), .device = (dev), \
	.subvendor = (svend), .subdevice = (sdev)
#endif

DEFINE_PCI_DEVICE_TABLE(ums7_id_table) = {
	{PCI_DEVICE_SUB(0x1556, 0x1103, 0x1556, 0x1103)},
	{PCI_DEVICE_SUB(0x1556, 0x1106, 0x10ee, 0x0007)},
	{PCI_DEVICE_SUB(0, 0, 0, 0)}
};
MODULE_DEVICE_TABLE(pci, ums7_id_table);

static struct pci_driver ums7_driver = {
	name: UMS7_CLASS,
	id_table: ums7_id_table,
	probe: ums7_probe,
	remove: ums7_remove,
};

static int ums7_init_module(void)
{
	return pci_register_driver(&ums7_driver);
}

static void ums7_cleanup_module(void)
{
	pci_unregister_driver(&ums7_driver);
}

module_init(ums7_init_module);
module_exit(ums7_cleanup_module);
